import re
import functools
from django.db import models


def length_of_string(s):
    return len(s) + functools.reduce(lambda x, y: x + len(y), re.findall("[\u2E80-\u9FFF]+", s), 0)


def update_instance(instance, data):
    empty = object()
    instance_list = [instance]
    for k, v in data.items():
        x = getattr(instance, k, empty)
        if x is not empty:
            if isinstance(x, models.Model):
                instance_list.extend(update_instance(x, v))
            else:
                setattr(instance, k, v)
    return instance_list



import sys
import functools
from django.db.models.fields.files import FieldFile
from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _
from django.contrib.admin.views.main import ChangeList
from ...utils.model import ModelFields


class MetaClass(admin.ModelAdmin.__class__):
    def __new__(mcs, name, bases, attrs):
        new_class = super(MetaClass, mcs).__new__(mcs, name, bases, attrs)
        if new_class._model_cls:
            admin.site.register(new_class._model_cls, new_class)
        return new_class


class ModelAdminAbstract(admin.ModelAdmin, metaclass=MetaClass):
    _model_cls = None


class ModelAdminBase(ModelAdminAbstract):
    _img_width = 50

    @property
    @functools.lru_cache(None)
    def _cl(self) -> ChangeList:
        count = 3  # only be used in custom field function, search for: lookup_field(field_name, result, cl.model_admin)
        while True:
            frame = sys._getframe(count)
            if 'cl' in frame.f_locals:
                return frame.f_locals['cl']
            count += 1

    @property
    @functools.lru_cache(None)
    def _results_index(self) -> dict:
        return {obj.pk: index for index, obj in enumerate(self._cl.result_list)}

    def _get_obj_index(self, obj):
        return (self._cl.page_num - 1) * self._cl.list_per_page + self._results_index[obj.pk]

    def row_index_(self, obj):
        return self._get_obj_index(obj) + 1

    row_index_.short_description = _("序号")

    @staticmethod
    def _get_url(obj):
        if isinstance(obj, FieldFile):
            return obj.url
        return obj

    @classmethod
    def format_self(cls, x):
        return x

    @classmethod
    def format_img(cls, img):
        if img:
            img = cls._get_url(img)
            return format_html('<img src="{0}" width="{1}" height="{1}"/>'.format(img, cls._img_width))
        else:
            return ""

    @classmethod
    def format_aimg(cls, img):
        if img:
            img = cls._get_url(img)
            return format_html(
                '<a href="{0}" target="_blank"><img src="{0}" width="{1}" height="{1}"/></a>'.
                format(img, cls._img_width))
        else:
            return ""

    @classmethod
    def format_imgs(cls, img_list):
        return format_html("".join([cls.format_img(x) for x in img_list]))

    @classmethod
    def format_aimgs(cls, img_list):
        return format_html("".join([cls.format_aimg(x) for x in img_list]))

    @staticmethod
    def format_tags(tags, ml=None, sep=' '):
        if ml is not None:
            tags = sorted(tags, key=lambda x: len(x))
            r = ""
            for tag in tags:
                temp = ""
                if r:
                    temp += sep
                temp += tag
                if len(r) + len(temp) > ml:
                    return r
                r += temp
            return r
        return sep.join(tags)

    @classmethod
    def format_a(cls, url, text=None, target=None):
        if url:
            url = cls._get_url(url)
            target = target or '_blank'
            if callable(text):
                text = text(url)
            elif text == 0:
                text = url
            elif text is None:
                text = url.rsplit('/', 1)[-1]
            return format_html("<a href='{url}' target='{target}'>{text}</a>", url=url, text=text, target=target)
        return ""


def calc_list_display(_model_cls, rewrite_set=None, disable_set=None, rewrite_suffix='_'):
    mfs = ModelFields(_model_cls)
    array = [] if mfs.pk in mfs.auto else [ModelAdminBase.row_index_.__name__]
    for name in mfs.sort_fields(mfs.auto.keys(), mfs.common.keys(), mfs.o2o.keys(), mfs.o2m.keys()):
        if disable_set and name in disable_set:
            continue
        if rewrite_set and name in rewrite_set:
            name += rewrite_suffix
        array.append(name)
    return tuple(array)


def calc_search_fields(_model_cls, list_display=None, disable_set=None):
    array = []
    mfs = ModelFields(_model_cls)
    for name in mfs.sort_fields(mfs.auto.keys(), mfs.common.keys(), mfs.o2o.keys(), mfs.o2m.keys()):
        if disable_set and name in disable_set:
            continue
        array.append(name)
    return tuple(item for item in array if item in list_display) if list_display else tuple(array)


def admin_register(_model_cls, list_display=None):
    list_display = list_display or calc_list_display(_model_cls)
    type(
        'admin',
        (ModelAdminBase,),
        {
            '_model_cls': _model_cls,
            'list_display': list_display,
            'search_fields': calc_search_fields(_model_cls, list_display)
        }
    )


def admin_register_batch(_model_cls_list):
    for _model_cls in _model_cls_list:
        admin_register(_model_cls)



from django.conf import settings
from django_typer import TyperCommand


class CommandBasic(TyperCommand):
    def just_for_debug_mode(self):
        if not settings.DEBUG:
            self.stdout.write("Only work on debug mode!")
            return True



import time
import logging
import datetime
from django.utils import timezone
from django.conf import settings
from dektools.time import TZ_UTC


class SnowflakeException(Exception):
    pass


class SnowflakeWorkerException(SnowflakeException):
    pass


class SnowflakeBackwardsException(SnowflakeException):
    pass


class SnowflakeGenerator:
    STEP_THRESHOLD_MAX_AMOUNT = 100  # 时钟回拨 个数
    STEP_THRESHOLD_MS = 128 * 2  # 时钟回拨 毫秒 https://docs.ntpsec.org/latest/clock.html
    EPOCH_TIMESTAMP = int(datetime.datetime(2022, 4, 11, 17, 15, 46).replace(tzinfo=TZ_UTC).timestamp() * 1000)
    END_TIMESTAMP = int(datetime.datetime(3022, 4, 11, 17, 15, 46).replace(tzinfo=TZ_UTC).timestamp() * 1000)

    BITS = ['dynamic', 15, 12]
    MAX_VALUE = 2 ** sum(BITS[1:]) - 1 + (END_TIMESTAMP - EPOCH_TIMESTAMP) * 2 ** sum(BITS[1:])
    MAX_VALUE_WORKER = 2 ** BITS[1] - 1

    @classmethod
    def new_instance(cls):
        return cls(int(settings.DEKDJTOOLS_SNOWFLAKE_INSTANCE))

    def __init__(self, worker):
        self.worker = worker & self.MAX_VALUE_WORKER
        if self.worker != worker:
            raise SnowflakeWorkerException('Worker wrong! %d > %d' % (worker, self.MAX_VALUE_WORKER))
        self.last_timestamp = self.EPOCH_TIMESTAMP
        self.sequence = 0
        self.sequence_overload = 0
        self.errors = 0
        self.generated_ids = 0
        self.step_threshold_list = []
        self.step_threshold_map = {}

    def save_step_threshold(self):
        if self.last_timestamp not in self.step_threshold_map:
            index = len(self.step_threshold_list)
            for i, ts in enumerate(reversed(self.step_threshold_list)):
                if self.last_timestamp > ts:
                    index = index - i - 1
                    break
            self.step_threshold_list.insert(index, self.last_timestamp)
        self.step_threshold_map[self.last_timestamp] = max(
            self.step_threshold_map.get(self.last_timestamp, 0),
            self.sequence
        )
        while len(self.step_threshold_list) > self.STEP_THRESHOLD_MAX_AMOUNT:
            if self.last_timestamp - self.step_threshold_list[0] > self.STEP_THRESHOLD_MS:
                ts = self.step_threshold_list.pop(0)
                del self.step_threshold_map[ts]
            else:
                break

    def load_step_threshold(self, curr_time):
        if not self.step_threshold_list:
            return False
        if self.step_threshold_list[0] > curr_time:
            return False
        sequence = self.step_threshold_map.get(curr_time, None)
        self.sequence = sequence if sequence is not None else 0
        self.last_timestamp = curr_time
        return True

    def get_next_id(self):
        curr_time = int(timezone.now().timestamp() * 1000)

        if curr_time < self.last_timestamp:
            if not self.load_step_threshold(curr_time):
                # stop handling requests til we've caught back up
                self.errors += 1
                raise SnowflakeBackwardsException('Clock went backwards! %d < %d' % (curr_time, self.last_timestamp))

        if curr_time > self.last_timestamp:
            self.sequence = 0
            self.last_timestamp = curr_time

        self.sequence += 1

        if self.sequence >= 2 ** self.BITS[2]:
            # the sequence is overload, just wait to next sequence
            logging.warning('The sequence has been overload')
            self.sequence_overload += 1
            time.sleep(0.001)
            return self.get_next_id()

        generated_id = ((curr_time - self.EPOCH_TIMESTAMP) << (self.BITS[1] + self.BITS[2])) \
                       | (self.worker << self.BITS[2]) \
                       | self.sequence

        self.generated_ids += 1
        self.save_step_threshold()
        return generated_id

    @property
    def stats(self):
        return {
            'worker': self.worker,
            'timestamp': int(timezone.now().timestamp() * 1000),  # current timestamp for this worker
            'last_timestamp': self.last_timestamp,  # the last timestamp that generated ID on
            'sequence': self.sequence,  # the sequence number for last timestamp
            'sequence_overload': self.sequence_overload,  # the number of times that the sequence is overflow
            'errors': self.errors,  # the number of times that clock went backward
        }


snowflake_generator = SnowflakeGenerator.new_instance()



import functools
from collections import OrderedDict
from itertools import chain
from django.db.models.fields.related import ForeignKey, OneToOneField, ManyToOneRel, ManyToManyRel, OneToOneRel
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation, GenericRel


class ModelFields:
    def __init__(self, model_cls):
        self.model_cls = model_cls
        self.pk = model_cls._meta.pk.name
        self.common = OrderedDict()
        self.auto = OrderedDict()
        self.ct = OrderedDict()
        self.ct_pk = OrderedDict()
        self.ct_query = OrderedDict()
        self.o2o = OrderedDict()
        self.o2m = OrderedDict()
        self.m2m = OrderedDict()
        self.o2o_r = OrderedDict()
        self.m2m_r = OrderedDict()
        self.m2o_r = OrderedDict()
        self.o2m_r = OrderedDict()
        self.attname = OrderedDict()
        self.parse()

    def parse(self):
        ct__ct_field = {}
        ct__fk_field = set()
        for field in self.model_cls._meta.private_fields:
            if isinstance(field, GenericForeignKey):
                ct__ct_field[field.ct_field] = field.name
                ct__fk_field.add(field.fk_field)
                self.ct[field.name] = None
                self.ct_pk[field.name] = field.fk_field
                self.attname[field.name] = field.fk_field
            elif isinstance(field, GenericRelation):
                self.o2m_r[field.name] = field.related_model
            else:
                raise LookupError
        for field in self.model_cls._meta.related_objects:
            if isinstance(field, ManyToOneRel):
                self.m2o_r[field.get_accessor_name()] = field.related_model
            elif isinstance(field, ManyToManyRel):
                self.m2m_r[field.get_accessor_name()] = field.related_model
            elif isinstance(field, OneToOneRel):
                self.o2o_r[field.get_accessor_name()] = field.related_model
            elif isinstance(field, GenericRel):
                gr = ct__ct_field[field.remote_field.content_type_field_name]
                self.ct[gr] = field.related_model
                self.ct_query[gr] = field.name
                self.attname[field.name] = self.attname.pop(gr)
            else:
                raise LookupError
        for field in self.model_cls._meta.many_to_many:
            self.m2m[field.name] = field.related_model
        for field in self.model_cls._meta.fields:
            if field.name in ct__fk_field or field.name in ct__ct_field:
                continue
            if field.auto_created:
                self.auto[field.name] = field
            elif isinstance(field, ForeignKey):
                self.o2m[field.name] = field.related_model
            elif isinstance(field, OneToOneField):
                self.o2o[field.name] = field.related_model
            else:
                self.common[field.name] = field
            self.attname[field.name] = field.attname

    @property
    @functools.lru_cache(None)
    def attrs(self):
        return set(self.common) | set(self.auto) | set(self.ct) | set(self.o2o) | set(self.o2m) | set(
            self.m2m) | set(self.o2o_r) | set(self.o2m_r) | set(self.m2m_r) | set(self.m2o_r)

    def sort_fields(self, *fields):
        all_fields = {field.name: i for i, field in enumerate(self.model_cls._meta.fields)}
        return sorted(chain(*fields), key=lambda x: all_fields[x])


def get_model_field_attr(model_cls, field_name: str, attr: str):
    return getattr(model_cls._meta.get_field(field_name), attr)



from django.conf import settings
from dektools.file import sure_dir
from ..base import CommandBasic
from .utils import get_backup_location


class Command(CommandBasic):
    def handle(self):
        location = get_backup_location()
        if location:
            sure_dir(location)
        sure_dir(settings.MEDIA_ROOT)



from django import forms
from django.db import models
from django.core import validators
from django.utils.translation import ugettext_lazy as _


class URLTextField(models.TextField):
    default_validators = [validators.URLValidator()]
    description = _("URL")

    def formfield(self, **kwargs):
        return super().formfield(**{
            'form_class': forms.URLField,
            **kwargs,
        })



from importlib import import_module
from django.apps import apps
from django.db import models


class DjFinder:
    def __init__(self, includes=None, excludes=None):
        self.includes = includes
        self.excludes = excludes

    def list_apps(self):
        for app_config in apps.get_app_configs():
            if self.excludes is not None:
                if app_config.name not in self.excludes:
                    yield app_config
            elif self.includes is not None:
                if app_config.name in self.includes:
                    yield app_config
            else:
                yield app_config

    def list_apps_part(self, name):
        for app in self.list_apps():
            try:
                yield app.label, import_module(f'{app.name}.{name}')
            except ModuleNotFoundError:
                pass

    def list_apps_part_cls(self, name, cls, *excludes):
        for a, xx in self.list_apps_part(name):
            for x in vars(xx).values():
                if x is not cls and isinstance(x, type) and issubclass(x, cls) and not issubclass(x, tuple(excludes)):
                    yield x, a

    def list_models(self):
        def sure_weight(x):
            max_value = 0
            for xx in x.mro():
                if xx is not x and xx in ms:
                    if xx not in weights:
                        weights[xx] = sure_weight(xx)
                    if weights[xx] > max_value:
                        max_value = weights[xx]
            w = max_value + delta
            weights[x] = w
            return w

        weights = {}
        ms = {x for x, _ in self.list_apps_part_cls('models', models.Model)}
        delta = (len(ms) + 1) * 10
        for m in ms:
            if m not in weights:
                weights[m] = sure_weight(m)

        return sorted(ms, key=lambda x: (sure_weight(x), m._meta.label))



from django.db import models
from django.db.models import Q, F
from django.utils.dateparse import parse_date


def filter_apply(model_cls, data, init=None, disable=None):
    result = Q()
    if data:
        for item in data:
            query = Q()
            for key, value in item.items():
                keys = key.split('__')
                _valid_path(keys, disable)
                field_cls = _resolve_field(model_cls, keys)
                if field_cls is None:
                    raise TypeError(keys)
                if isinstance(value, list):
                    if value[0] is not None:
                        query &= Q(**{f'{key}__gte': _format_value(field_cls, value[0])})
                    if value[1] is not None:
                        query &= Q(**{f'{key}__lte': _format_value(field_cls, value[1])})
                else:
                    if isinstance(field_cls, (models.CharField, models.TextField)):
                        key = f'{key}__icontains'
                    query &= Q(**{key: _format_value(field_cls, value)})
            result |= query
    if init:
        if not isinstance(init, Q):
            init = Q(**init)
        result &= init
    return result


def order_apply(data, disable=None):
    if disable and data:
        for key in data:
            _valid_path(key.strip('$^-').split('__'), disable)
    return parse_order(data)


def parse_order(array):
    result = []
    for item in array:
        if item.startswith('-'):
            item = item[1:]
            if item.startswith('$'):
                v = F(item[1:]).desc(nulls_last=True)
            elif item.startswith('^'):
                v = F(item[1:]).desc(nulls_first=True)
            else:
                v = F(item).desc()
        else:
            if item.startswith('$'):
                v = F(item[1:]).asc(nulls_last=True)
            elif item.startswith('^'):
                v = F(item[1:]).asc(nulls_first=True)
            else:
                v = F(item).asc()
        result.append(v)
    return result


def _valid_path(paths, disable):
    if not disable:
        return
    cursor = disable
    for path in paths:
        cursor = cursor.get(path)
        if cursor is False:
            raise KeyError(paths)
        if cursor is None:
            return


def _resolve_field(model_cls, paths):
    cursor = model_cls
    for path in paths:
        if issubclass(cursor, models.Model):
            cursor = cursor._meta.get_field(path)
            rm = cursor.related_model
            if rm:
                cursor = rm
        else:
            return None
    return cursor


def _format_value(field, value):
    if isinstance(field, (models.DateField, models.DateTimeField, models.TimeField)):
        return parse_date(value)
    else:
        return value



from django.contrib.auth import get_user_model


def get_user_obj(user):
    if not user.pk:
        return None
    um = get_user_model()
    if isinstance(user, um):
        return user
    else:
        try:
            return um.objects.get(pk=user.pk)
        except um.DoesNotExist:
            return None


def get_user_ip(request):
    x = request.META.get('HTTP_X_FORWARDED_FOR')
    return x if x else request.META['REMOTE_ADDR']



import sys
from ..base import CommandBasic
from .utils import get_backup_location


class Command(CommandBasic):
    def handle(self):
        location = get_backup_location()
        if location:
            sys.stdout.write(location)
            sys.stdout.flush()



import os

default_db_options = {
    "ENGINE": "psqlextra.backend",
    # 'ENGINE': 'django.db.backends.postgresql',
    'NAME': "default",
    'USER': "postgres",
    'PASSWORD': "123456",
    'HOST': 'localhost',
    'PORT': 5432,
}
default_cache_options = {
    "BACKEND": "django_redis.cache.RedisCache",
    "LOCATION": "redis://:%s@%s:%s/%s",
    "OPTIONS": {
        "CLIENT_CLASS": "django_redis.client.DefaultClient",
    }
}


def generate_databases(s):
    result = {}
    if s:
        db_name, s = s.split('::')
        if db_name:
            db_name_options = {'NAME': db_name}
        else:
            db_name_options = {}
        for item in s.split(','):
            item = item.strip()
            if not item:
                continue
            name, host = item.split(':')
            result[name] = {**default_db_options, **{'HOST': host}, **db_name_options}
    return result


def generate_caches(s):
    result = {}
    if s:
        for item in s.split(','):
            item = item.strip()
            if not item:
                continue
            name, host, index = item.split(':')
            result[name] = {
                **default_cache_options,
                **{
                    'LOCATION': default_cache_options['LOCATION'] % (
                        "123456", host, 6379, index or 0,
                    )
                }
            }
    return result


extra_databases = generate_databases(os.getenv("DJANGO_PRESET_DATABASES", ""))
extra_caches = generate_caches(os.getenv("DJANGO_PRESET_CACHES", ""))


def extra_fix(installed_apps):
    if not any(x for x in extra_databases.values() if "psqlextra" in x["ENGINE"]):
        try:
            installed_apps.remove('psqlextra')
        except ValueError:
            pass



from django.db import models
from django.utils.translation import ugettext_lazy as _
from ..snowflake import SnowflakeGenerator


class SnowflakeField(models.PositiveBigIntegerField):
    description = _("雪花ID")
    MAX_VALUE = SnowflakeGenerator.MAX_VALUE

    def __init__(self, **kwargs):
        self.snowflake_generator = SnowflakeGenerator.new_instance()
        kwargs.update(dict(
            editable=False,
            blank=True
        ))
        super().__init__(**kwargs)

    def pre_save(self, model_instance, add):
        if add:
            value = self.snowflake_generator.get_next_id()
            setattr(model_instance, self.attname, value)
            return value
        else:
            return super().pre_save(model_instance, add)

    def formfield(self, **kwargs):
        return super().formfield(**{
            'min_value': -SnowflakeField.MAX_VALUE - 1,
            'max_value': SnowflakeField.MAX_VALUE,
            **kwargs,
        })



import functools
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils.functional import classproperty
from django.contrib.contenttypes.fields import GenericForeignKey
from django.core.exceptions import FieldDoesNotExist
from ...utils.format import model_to_json, model_to_data


class RelatedRef:
    def __init__(self, instance):
        self.instance = instance

    def __getattr__(self, item):
        if hasattr(self.instance, item):
            return getattr(self.instance, item)
        else:
            return None


class ManagerBasic(models.Manager):
    pass


class ModelBasic(models.Model):
    objects = ManagerBasic()

    class Meta:
        abstract = True

    def format_as_json(self, **kwargs):
        return model_to_json(self, **kwargs)

    def format_as_data(self):
        return model_to_data(self)

    @property
    def related_(self):
        return RelatedRef(self)

    @classmethod
    def has_field(cls, name):
        try:
            cls._meta.get_field(name)
            return True
        except FieldDoesNotExist:
            return False

    @classproperty
    @functools.lru_cache(None)
    def fields_local_keys(self):
        result = set()
        for field in self._meta.local_fields:
            if not field.auto_created:
                result.add(field.name)
        return result

    @classproperty
    @functools.lru_cache(None)
    def fields_editable_keys(self):
        result = set()

        for field in self._meta.local_fields:
            if field.auto_created or getattr(field, 'auto_now_add', False) or getattr(field, 'auto_now', False):
                continue
            result.add(field.name)

        for field in self._meta.get_fields():
            if field.is_relation and isinstance(field, GenericForeignKey):
                result.add(field.name)
        return result

    @classmethod
    def fields_kwargs_filter(cls, kwargs):
        return {k: v for k, v in kwargs.items() if k in cls.fields_local_keys}


class DateTimeModel(ModelBasic):
    datetime_created = models.DateTimeField(_('创建时间'), auto_now_add=True)
    datetime_modified = models.DateTimeField(_('更新时间'), auto_now=True)

    class Meta:
        abstract = True


class StubManager(ManagerBasic):
    def stub(self):
        return self.get_queryset().filter(is_deleted=False)


class StubModel(ModelBasic):
    is_deleted = models.BooleanField(_('是否已被删除'), default=False)

    objects = StubManager()

    class Meta:
        abstract = True


class BaseModel(DateTimeModel, StubModel):
    class Meta:
        abstract = True



import os
import getpass
import typer
from typing_extensions import Annotated
from django.contrib.auth import get_user_model
from ..base import CommandBasic

User = get_user_model()


class Command(CommandBasic):
    help = "Ensure creating an admin user"

    def handle(
            self, username='', email='', password='',
            _input: Annotated[bool, typer.Option("--input/--no-input")] = False):
        if not _input:
            username = os.environ['DJANGO_SUPERUSER_USERNAME']
            email = os.environ['DJANGO_SUPERUSER_EMAIL']
            password = os.environ['DJANGO_SUPERUSER_PASSWORD']
        elif not password:
            password = getpass.getpass('Password: ')
        try:
            user = User.objects.get(**{User.USERNAME_FIELD: username})
            user.email = email
            user.set_password(password)
            user.save()
        except User.DoesNotExist:
            User.objects.create_superuser(**{
                User.USERNAME_FIELD: username,
                'email': email,
                'password': password
            })
